"""Datasources endpoints namespace"""

import sys
from datetime import datetime
from functools import partial
from typing import Optional

from ..dtos import MetisDataSourceContentOnlyDTO, MetisDataSourceDTO, MetisRequestIdDTO
from ..helpers import raise_on_metis_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence


class MetisDatasourcesNamespace(BaseNamespace):
    """Datasources endpoints namespace"""

    async def create_event(self, content: str) -> MetisRequestIdDTO:
        "Create data source"
        async with self._client.request(
            method="POST",
            url=self._base_url,
            json={"content": content},
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def create(self, content: str) -> Optional[MetisDataSourceDTO]:
        "Create data source and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.create_event, content)
        )
        if evt["type"] == "datasources":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("createdAt", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    async def delete_event(self, data_id: int) -> MetisRequestIdDTO:
        "Delete data source by id"
        async with self._client.request(
            method="DELETE",
            url=self._base_url / str(data_id),
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def delete(self, data_id: int) -> None:
        "Delete data source by id and wait for result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.delete_event, data_id)
        )

    async def list_event(self) -> MetisRequestIdDTO:
        "List data sources"
        async with self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def list(self) -> Sequence[MetisDataSourceDTO]:
        "List data sources and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "datasources":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover

    @raise_on_metis_error
    async def get(self, data_id: int) -> MetisDataSourceContentOnlyDTO:
        "Get data source by id"
        async with self._client.request(
            method="GET",
            url=self._base_url / str(data_id),
            auth_required=True,
        ) as resp:
            return await resp.json()



"""Metis API client"""

from .const import PROJECT_VERSION as __version__
from .dtos import MetisErrorDTO
from .metis import MetisAPI
from .metis_async import MetisAPIAsync
from .models import MetisLocalUserAuth, MetisNoAuth, MetisTokenAuth



"""Error DTOs"""

import sys
from typing import Union

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class MetisErrorErrorDTO(TypedDict):
    "Error's error payload DTO"
    message: str


class MetisErrorDTO(TypedDict):
    "Error DTO"
    status: int
    error: Union[MetisErrorErrorDTO, str]



"""DTOs"""

from .auth import MetisAuthCredentialsRequestDTO
from .base import MetisTimestampsDTO
from .calculation import MetisCalculationDTO
from .collection import (
    MetisCollectionCreateDTO,
    MetisCollectionDTO,
    MetisCollectionTypeDTO,
    MetisCollectionVisibility,
)
from .datasource import MetisDataSourceContentOnlyDTO, MetisDataSourceDTO
from .error import MetisErrorDTO
from .event import (
    MetisCalculationsEventDTO,
    MetisCollectionsEventDTO,
    MetisDataSourcesEventDTO,
    MetisErrorEventDTO,
    MetisEventDTO,
    MetisEventType,
    MetisPongEventDTO,
)
from .resp import MetisRequestIdDTO
from .user import MetisUserDTO



"""Small helpers"""

import re
import sys
from contextlib import suppress
from copy import deepcopy
from datetime import datetime
from functools import partial, wraps
from typing import Any, Optional, Type, TypeVar, cast

from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPForbidden,
    HTTPMisdirectedRequest,
    HTTPNotFound,
    HTTPPaymentRequired,
    HTTPUnauthorized,
)

from .dtos import MetisErrorDTO, MetisErrorEventDTO, MetisEventDTO
from .exc import (
    MetisAuthenticationException,
    MetisError,
    MetisNotFoundException,
    MetisPayloadException,
    MetisQuotaException,
)

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Mapping, Sequence, Tuple
else:  # pragma: no cover
    from collections.abc import Mapping, Sequence

    Tuple = tuple

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import TypeGuard
else:  # pragma: no cover
    from typing import TypeGuard


LAST_Z_REGEX = re.compile(r"Z$", re.IGNORECASE)


def parse_rfc3339(dt_str: Any) -> Optional[datetime]:
    "Parse RFC 3339 date string to datetime object"
    if isinstance(dt_str, datetime):
        return dt_str

    if not dt_str:
        return None

    if dt_str.endswith("Z"):
        dt_str = LAST_Z_REGEX.sub("+00:00", dt_str)

    with suppress(ValueError):
        datetime.fromisoformat(dt_str)

    for fmt in (
        "%Y-%m-%dT%H:%M:%S.%f%z",
        "%Y-%m-%d %H:%M:%S.%f%z",
        "%Y-%m-%dT%H:%M:%S%z",
        "%Y-%m-%d %H:%M:%S%z",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
    ):
        with suppress(ValueError):
            return datetime.strptime(dt_str, fmt)


_DT = TypeVar("_DT", bound=Mapping)


def convert_dict_items_to_datetime(keys: Sequence[str], data: _DT) -> _DT:
    "Convert mapping's items to datetime"

    def apply_parse(x: Tuple):
        if x[0] not in keys or not isinstance(x[1], str):
            return x
        return (x[0], parse_rfc3339(x[1]))

    return cast(_DT, dict(map(apply_parse, deepcopy(data).items())))


dict_dt_from_dt_str = partial(
    convert_dict_items_to_datetime, ["createdAt", "updatedAt"]
)


def is_metis_errors_evt_dto(evt: MetisEventDTO) -> TypeGuard[MetisErrorEventDTO]:
    "MetisEventDTO type guard"
    return bool(evt["type"] == "errors" and evt.get("data"))


def is_metis_error_dto(something) -> TypeGuard[MetisErrorDTO]:
    "MetisErrorDTO type guard"
    return (
        isinstance(something, dict)
        and "status" in something
        and "error" in something
        and (
            isinstance(something["error"], str)
            or isinstance(something["error"], dict)
            and "message" in something["error"]
            and something["error"]["message"]
        )
    )


def http_to_metis_error_map(status: int) -> Type[MetisError]:
    "Map HTTP exception to MetisError"
    err = MetisError
    if status in (HTTPForbidden.status_code, HTTPUnauthorized.status_code):
        err = MetisAuthenticationException
    if status == HTTPNotFound.status_code:
        err = MetisNotFoundException
    if status == HTTPBadRequest.status_code:
        err = MetisPayloadException
    if status == HTTPPaymentRequired.status_code:
        err = MetisQuotaException
    if status == HTTPMisdirectedRequest.status_code:
        err = MetisError
    return err


def metis_error_to_raise(dto: MetisErrorDTO):
    "Raise MetisErrorDTO"
    if isinstance(dto["error"], dict):
        msg = str(dto["error"]["message"])
    else:
        msg = dto["error"]
    err = http_to_metis_error_map(dto["status"])
    raise err(status=dto["status"], message=msg)


def raise_on_metis_error(func):
    "Raise on MetisErrorDTO in result"

    @wraps(func)
    async def wrapped(*args, **kwargs):
        result = await func(*args, **kwargs)
        if is_metis_error_dto(result):
            metis_error_to_raise(result)

        return result

    return wrapped


def raise_on_metis_error_in_event(func):
    "Raise on MetisErrorDTO in MetisEventDTO"

    @wraps(func)
    async def wrapped(*args, **kwargs):
        result = await func(*args, **kwargs)
        if is_metis_errors_evt_dto(result):
            errors = result.get("data", {}).get("data", [])
            if errors:
                metis_error_to_raise(errors[-1])

        return result

    return wrapped



"""Models"""

from .auth import BaseAuthenticator, MetisLocalUserAuth, MetisNoAuth, MetisTokenAuth
from .base import MetisBase
from .event import MetisMessageEvent
from .hub import MetisHub
from .subscription import MetisSubscription, act_and_get_result_from_stream



"""Low level http and SSE client"""

import sys
from asyncio import CancelledError, TimeoutError as AsyncioTimeoutError, sleep
from concurrent.futures import TimeoutError as FuturesTimeoutError
from datetime import timedelta
from json import JSONDecodeError
from typing import Any, Optional, Union

import aiohttp
from aiohttp import ClientResponse, ClientTimeout, RequestInfo
from aiohttp.client import _RequestContextManager
from aiohttp.client_exceptions import (
    ClientConnectionError,
    ClientPayloadError,
    ClientResponseError,
)
from aiohttp.web_exceptions import (
    HTTPInternalServerError,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from aiohttp_sse_client import client as sse_client
from yarl import URL

from .const import HttpMethods
from .exc import MetisConnectionException, MetisException
from .helpers import http_to_metis_error_map
from .models import BaseAuthenticator, MetisBase, MetisNoAuth

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Callable, Dict, Mapping
else:  # pragma: no cover
    from collections.abc import Callable, Mapping

    Dict = dict

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired, TypedDict, Unpack
else:  # pragma: no cover
    from typing import NotRequired, TypedDict, Unpack


class ClientRequestKwargs(TypedDict):
    "MetisClient.create kwargs"
    json: NotRequired[Any]
    data: NotRequired[Union[str, bytes]]
    headers: NotRequired[Mapping[str, Any]]
    method: NotRequired[HttpMethods]
    params: NotRequired[Dict[str, Any]]
    timeout: NotRequired[float]
    auth_required: NotRequired[bool]


class MetisClient(MetisBase):
    """
    Client to handle API calls.
    Don't use this directly, use `metis_client.metis.Metis` to get the client.
    """

    _auth: BaseAuthenticator
    _base_url: URL

    def __init__(
        self,
        session: aiohttp.ClientSession,
        base_url: URL,
        auth: Optional[BaseAuthenticator] = None,
    ) -> None:
        """
        Initialize the Metis API client.
        **Arguments:**
        `session`: The aiohttp client session to use for making requests.
        `base_url`: Root URL in form of `yarl.URL`.
        `auth`: Authenticator, subclass of `BaseAuthenticator`.
        """
        self._session = session
        self._auth = auth or MetisNoAuth()
        if not base_url.is_absolute():
            raise TypeError("Base URL should be absolute")
        self._base_url = base_url

    def _url_rel_to_abs(self, url: URL) -> URL:
        "If url is relative, join it with base url, else return as is"
        if url.is_absolute():
            return url
        if url.path.startswith("/"):
            return self._base_url.join(url)
        return (self._base_url / url.path).with_query(url.query)

    async def _do_auth(self, force: bool = False) -> None:
        async with self._auth.lock:
            if force or await self._auth.should_update(self._session, self._base_url):
                await self._auth.authenticate(self._session, self._base_url)

    @staticmethod
    def _raise_for_status(
        status: int, req_info: RequestInfo, msg: Optional[str] = None
    ) -> None:
        if status < 400:
            return
        msg = f"Response exception for {req_info.url!s} with code {status}: {msg!s}"
        proto_err = http_to_metis_error_map(status)
        raise proto_err(status=status, message=msg) from ClientResponseError(
            req_info, message=msg, status=status, history=()
        )

    async def _request(
        self, url: URL, **opts: Unpack[ClientRequestKwargs]
    ) -> ClientResponse:
        """
        Makes an HTTP request to the specified endpoint using the specified parameters.
        """
        url = self._url_rel_to_abs(url)
        method = opts.get("method", "GET")
        auth_required = opts.get("auth_required", False)
        aio_opts = {
            k: v
            for k, v in opts.items()
            if k in ["params", "timeout", "headers", "json", "data"]
        }
        aio_opts["raise_for_status"] = False

        # preauthenticate
        if auth_required:
            await self._do_auth()

        try:
            result = await self._session.request(method, url, **aio_opts)

            # redo if failed because of auth
            if result.status == HTTPUnauthorized.status_code:
                # forced auth for first request
                # normal auth (only if needed) for all other
                await self._do_auth(force=not self._auth.lock.locked())
                result.close()
                result = await self._session.request(method, url, **aio_opts)
            # rate limit - redo all
            if result.status == HTTPTooManyRequests.status_code:
                await sleep(10)
                result.close()
                return await self._request(url, **opts)

        except (CancelledError, ClientConnectionError) as exc:
            raise MetisConnectionException(
                f"Request exception for {str(url)!r} with - {exc}"
            ) from exc

        except (TimeoutError, AsyncioTimeoutError, FuturesTimeoutError):
            raise MetisConnectionException(
                f"Timeout of {opts.get('timeout')} reached while waiting for {str(url)}"
            ) from None

        except BaseException as exc:  # pragma: no cover
            raise MetisException(
                f"Unexpected exception for {str(url)!r} with - {exc}"
            ) from exc

        msg = None
        try:
            body = None
            if result.content_type.startswith("application/json"):
                body = await result.json(
                    encoding="utf-8", content_type=result.content_type
                )
                if isinstance(body, dict) and body.get("error", None):
                    msg = str(body.get("error"))
            elif result.content_type.startswith("text/"):
                msg = await result.text("utf-8")
            else:
                msg = str(await result.read())
        except (ClientPayloadError, JSONDecodeError) as exc:
            raise MetisException(
                f"Broken payload data from {str(url)!r}: {exc}"
            ) from exc
        except BaseException as exc:  # pragma: no cover
            raise MetisException(
                f"Could not handle response data from {str(url)!r} with - {exc}"
            ) from exc

        self._raise_for_status(result.status, result.request_info, msg)

        return result

    # pylint: disable=too-many-arguments
    def request(
        self,
        url: URL,
        **opts: Unpack[ClientRequestKwargs],
    ) -> _RequestContextManager:
        """
        Makes an HTTP request to the specified endpoint using the specified parameters.
        This method is asynchronous, meaning that it will not block the execution
        of the program while the request is being made and processed.
        **Arguments**:
        -  `url` (Required): The API endpoint to call.
        **Optional arguments**:
        - `data`: The data to include in the request body. Can be a dictionary,
           a string, or None.
        - `headers`: The headers to include in the request. Can be a dictionary or None.
        - `method`: The HTTP method to use for the request. Defaults to GET.
        - `params`: The query parameters to include in the request.
           Can be a dictionary or None.
        - `timeout`: The maximum amount of time to wait for the request to complete,
          in seconds. Can be an integer or None.
        - `auth_required`: Flag that auth requered for this request
        Returns:
        A `aiohttp.client._RequestContextManager` object representing the API response
        """
        return _RequestContextManager(self._request(url, **opts))

    async def sse(
        self,
        url: URL,
        on_message: Callable[[sse_client.MessageEvent], None],
        on_open: Optional[Callable[[], None]] = None,
        params: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
    ) -> None:
        """
        Creates `aiohttp_sse_client.client.EventSource` object with sensible defaults.
        **Arguments**:
        - `url` (Required): The API endpoint to connect.
        - `on_message`: Message callback
        **Optional arguments**:
        - `on_open`: Callback when connected
        - `params`: The query parameters to include in the request.
           Can be a dictionary or None.
        - `timeout`: Stream timeout, None for infinity
        Returns: None
        """
        url = self._url_rel_to_abs(url)
        original_backoff = 0.1
        backoff = original_backoff

        es_timeout = ClientTimeout(total=timeout, sock_connect=600, sock_read=None)
        while True:
            try:
                if backoff > original_backoff:
                    await sleep(backoff)
                await self._do_auth()
                async with sse_client.EventSource(
                    str(url),
                    reconnection_time=timedelta(seconds=original_backoff),
                    max_connect_retry=0,
                    on_open=on_open,
                    session=self._session,
                    params=params,
                    timeout=es_timeout,
                    read_bufsize=2**19,
                    raise_for_status=True,
                ) as evt_src:
                    backoff = original_backoff
                    async for evt in evt_src:
                        on_message(evt)

            except (TimeoutError, AsyncioTimeoutError, FuturesTimeoutError):
                backoff *= 1.5
                self.logger.warning(
                    "%s endpoint timeouted - reconnecting", self._base_url
                )
                continue

            except CancelledError:
                # task is cancelled - abort
                break

            except (ClientConnectionError, ConnectionError) as exc:
                raise MetisConnectionException(
                    f"Connection error for {str(url)!r} with - {exc}"
                ) from exc

            except ClientResponseError as err:
                if err.status and (
                    err.status == HTTPTooManyRequests.status_code
                    or err.status >= HTTPInternalServerError.status_code
                ):
                    backoff *= 2
                    self.logger.warning(
                        "%s connection error %s - reconnecting in %s seconds",
                        url,
                        err,
                        backoff,
                    )
                    continue
                self._raise_for_status(err.status, err.request_info, err.message)



"""Authentication endpoints namespace"""

from ..dtos import MetisAuthCredentialsRequestDTO, MetisUserDTO
from ..helpers import dict_dt_from_dt_str, raise_on_metis_error
from .base import BaseNamespace


class MetisAuthNamespace(BaseNamespace):
    """Authentication endpoints namespace"""

    @raise_on_metis_error
    async def login(self, email: str, password: str) -> bool:
        "Login"
        payload = MetisAuthCredentialsRequestDTO(email=email, password=password)
        async with self._client.request(
            method="POST",
            url=self._base_url,
            json=payload,
        ) as resp:
            return resp.ok

    @raise_on_metis_error
    async def whoami(self) -> MetisUserDTO:
        "Get self info"
        async with self._client.request(
            url=self._base_url,
            auth_required=self._auth_required,
        ) as resp:
            return dict_dt_from_dt_str(await resp.json())



"""Calculation DTOs"""
import sys

from .base import MetisTimestampsDTO
from .datasource import MetisDataSourceDTO

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired
else:  # pragma: no cover
    from typing import NotRequired


class MetisCalculationDTO(MetisTimestampsDTO):
    "Calculation DTO"
    id: int
    name: str
    userId: int
    status: int
    progress: int
    result: NotRequired[Sequence[MetisDataSourceDTO]]



"""v0 namespace"""

from .auth import MetisAuthNamespace
from .base import BaseNamespace
from .calculations import MetisCalculationsNamespace
from .collections import MetisCollectionsNamespace
from .datasources import MetisDatasourcesNamespace


class MetisV0Namespace(BaseNamespace):
    """v0 namespace"""

    def __post_init__(self) -> None:
        self.__ns_auth = MetisAuthNamespace(
            self._client, self._base_url / "auth", root=self._root
        )
        self.__ns_calculations = MetisCalculationsNamespace(
            self._client, self._base_url / "calculations", root=self._root
        )
        self.__ns_collections = MetisCollectionsNamespace(
            self._client, self._base_url / "collections", root=self._root
        )
        self.__ns_datasources = MetisDatasourcesNamespace(
            self._client, self._base_url / "datasources", root=self._root
        )

    async def ping(self) -> None:
        "Run ping pong game"
        await self._client.request(url=self._base_url, method="HEAD", json={})

    @property
    def auth(self) -> MetisAuthNamespace:
        "Property to access the auth namespace."
        return self.__ns_auth

    @property
    def calculations(self) -> MetisCalculationsNamespace:
        "Property to access the calculations namespace."
        return self.__ns_calculations

    @property
    def collections(self) -> MetisCollectionsNamespace:
        "Property to access the collections namespace."
        return self.__ns_collections

    @property
    def datasources(self) -> MetisDatasourcesNamespace:
        "Property to access the datasources namespace."
        return self.__ns_datasources



"Stream hub"

from asyncio import Event
from typing import TYPE_CHECKING, Set

from ..dtos import MetisEventDTO
from .base import MetisBase

if TYPE_CHECKING:  # pragma: no cover
    from .subscription import MetisSubscription


class MetisHub(MetisBase):
    "Stream hub"
    _subscriptions: "Set[MetisSubscription]"

    def __init__(self) -> None:
        self._subscriptions = set()
        self._connected_event = Event()

    def __len__(self) -> int:
        return len(self._subscriptions)

    @property
    def connected(self) -> bool:
        "Check if connected event is set"
        return self._connected_event.is_set()

    def set_connected(self) -> None:
        "Set connected event"
        self._connected_event.set()

    def set_disconnected(self) -> None:
        "Clear connected event"
        self._connected_event.clear()

    async def wait_connected(self) -> None:
        "Wait connected event"
        await self._connected_event.wait()

    def subscribe(self, subscription: "MetisSubscription") -> None:
        "Register subscription"
        self._subscriptions.add(subscription)

    def unsubscribe(self, subscription: "MetisSubscription") -> None:
        "Unsubscribe subscription"
        self._subscriptions.discard(subscription)

    def unsubscribe_all(self) -> None:
        "Unsubscribe all subscriptions"
        self._subscriptions.clear()

    async def close(self) -> None:
        "Close all subscriptions"
        subs = list(self._subscriptions)
        self.unsubscribe_all()
        for sub in subs:
            await sub.close()

    def publish(self, evt: MetisEventDTO) -> None:
        "Publish message to subscriptions"
        for sub in self._subscriptions:
            sub.put_nowait(evt)



"""Base class for all GitHub objects."""

from logging import Logger

from ..const import LOGGER


# pylint: disable=too-few-public-methods
class MetisBase:
    """Base class for all Metis objects."""

    logger: Logger = LOGGER



"""Authentication DTOs"""

import sys

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class MetisAuthCredentialsRequestDTO(TypedDict):
    "Authentication request payload"
    email: str
    password: str



"""Collections endpoints namespace"""
import sys
from datetime import datetime
from functools import partial
from typing import Optional

from ..dtos import MetisCollectionCreateDTO, MetisCollectionDTO, MetisRequestIdDTO
from ..helpers import raise_on_metis_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired, TypedDict, Unpack
else:  # pragma: no cover
    from typing import NotRequired, TypedDict, Unpack


class MetisCollectionsCreateKwargs(TypedDict):
    "MetisCollectionsNamespace.create kwargs"
    description: NotRequired[str]
    data_source_ids: NotRequired[Sequence[int]]
    user_ids: NotRequired[Sequence[int]]


class MetisCollectionsNamespace(BaseNamespace):
    """Collections endpoints namespace"""

    async def create_event(
        self, type_id: int, title: str, **opts: Unpack[MetisCollectionsCreateKwargs]
    ) -> MetisRequestIdDTO:
        "Create collection"
        payload = MetisCollectionCreateDTO(
            title=title,
            typeId=type_id,
            description=opts.get("description", ""),
            dataSources=opts.get("data_source_ids", []),
            users=opts.get("user_ids", []),
        )
        async with await self._client.request(
            method="PUT",
            url=self._base_url,
            json=payload,
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def create(
        self, type_id: int, title: str, **opts: Unpack[MetisCollectionsCreateKwargs]
    ) -> Optional[MetisCollectionDTO]:
        "Create collection and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe,
            partial(self.create_event, type_id, title, **opts),
        )
        if evt["type"] == "collections":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("createdAt", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    async def list_event(self) -> MetisRequestIdDTO:
        "List user's collections by criteria"
        async with await self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json()

    @raise_on_metis_error
    async def list(self) -> Sequence[MetisCollectionDTO]:
        "List user's collections by criteria and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "collections":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover

    async def delete_event(self, collection_id: int) -> MetisRequestIdDTO:
        "Remove a collection"
        async with await self._client.request(
            method="DELETE",
            url=self._base_url / str(collection_id),
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def delete(self, collection_id: int) -> None:
        "Remove a collection by id and wait for result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.delete_event, collection_id)
        )



"""Metis API synchronous client"""

import sys
from functools import partial
from typing import Any, TypeVar, cast

from aiohttp.typedefs import StrOrURL
from asgiref.sync import async_to_sync

from .metis_async import MetisAPIAsync, MetisAPIKwargs
from .models.base import MetisBase
from .namespaces.collections import MetisCollectionsCreateKwargs

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Awaitable, Callable
else:  # pragma: no cover
    from collections.abc import Awaitable, Callable

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Concatenate, ParamSpec, Unpack
else:  # pragma: no cover
    from typing import Concatenate, ParamSpec, Unpack

AsyncClientGetter = Callable[[], MetisAPIAsync]
ReturnT_co = TypeVar("ReturnT_co", covariant=True)
ParamT = ParamSpec("ParamT")


# pylint: disable=too-few-public-methods
class MetisNamespaceSyncBase:
    "Base for synchronous namespaces"
    _client_getter: AsyncClientGetter

    def __init__(self, client_getter: AsyncClientGetter):
        self._client_getter = client_getter


def to_sync_with_metis_client(
    func: Callable[Concatenate[Any, MetisAPIAsync, ParamT], Awaitable[ReturnT_co]]
) -> Callable[Concatenate[Any, ParamT], ReturnT_co]:
    """
    Wraps async MetisNamespaceSync's method.
    - start MetisAPIAsync client,
    - pass it to the method as first argument after Self,
    - close client
    - wrap all with async_to_sync converter
    """

    async def inner(
        self: MetisNamespaceSyncBase, *args: ParamT.args, **kwargs: ParamT.kwargs
    ) -> ReturnT_co:
        # pylint: disable=protected-access
        async with self._client_getter() as client:
            return await func(self, client, *args, **kwargs)

    return cast(Any, async_to_sync(inner))


class MetisAuthNamespaceSync(MetisNamespaceSyncBase):
    """Authentication endpoints namespace"""

    @to_sync_with_metis_client
    async def login(self, client: MetisAPIAsync, email: str, password: str):
        "Login"
        return await client.v0.auth.login(email, password)

    @to_sync_with_metis_client
    async def whoami(self, client: MetisAPIAsync):
        "Get self info"
        return await client.v0.auth.whoami()


class MetisDatasourcesNamespaceSync(MetisNamespaceSyncBase):
    """Datasources endpoints namespace"""

    @to_sync_with_metis_client
    async def create(self, client: MetisAPIAsync, content: str):
        "Create data source and wait for result"
        return await client.v0.datasources.create(content)

    @to_sync_with_metis_client
    async def delete(self, client: MetisAPIAsync, data_id: int):
        "Delete data source by id and wait for result"
        return await client.v0.datasources.delete(data_id)

    @to_sync_with_metis_client
    async def list(self, client: MetisAPIAsync):
        "List data sources and wait for result"
        return await client.v0.datasources.list()

    @to_sync_with_metis_client
    async def get(self, client: MetisAPIAsync, data_id: int):
        "Get data source by id"
        return await client.v0.datasources.get(data_id)


class MetisCalculationsNamespaceSync(MetisNamespaceSyncBase):
    """Calculations endpoints namespace"""

    @to_sync_with_metis_client
    async def cancel(self, client: MetisAPIAsync, calc_id: int):
        "Cancel calculation and wait for result"
        return await client.v0.calculations.cancel(calc_id)

    @to_sync_with_metis_client
    async def create(self, client: MetisAPIAsync, data_id: int, engine: str = "dummy"):
        "Create calculation and wait for result"
        return await client.v0.calculations.create(data_id, engine)

    @to_sync_with_metis_client
    async def get_engines(self, client: MetisAPIAsync):
        "Get supported calculation engines"
        return await client.v0.calculations.get_engines()

    @to_sync_with_metis_client
    async def list(self, client: MetisAPIAsync):
        "List all user's calculations and wait for result"
        return await client.v0.calculations.list()


class MetisCollectionsNamespaceSync(MetisNamespaceSyncBase):
    """Collections endpoints namespace"""

    @to_sync_with_metis_client
    async def create(
        self,
        client: MetisAPIAsync,
        type_id: int,
        title: str,
        **opts: Unpack[MetisCollectionsCreateKwargs],
    ):
        "Create collection and wait for result"
        return await client.v0.collections.create(type_id, title, **opts)

    @to_sync_with_metis_client
    async def list(self, client: MetisAPIAsync):
        "List user's collections by criteria and wait for result"
        return await client.v0.collections.list()

    @to_sync_with_metis_client
    async def delete(self, client: MetisAPIAsync, collection_id: int):
        "Remove a collection by id and wait for result"
        return await client.v0.collections.delete(collection_id)


# pylint: disable=too-few-public-methods
class MetisV0NamespaceSync(MetisNamespaceSyncBase):
    """v0 namespace"""

    def __init__(self, client_getter: AsyncClientGetter):
        super().__init__(client_getter)
        self.auth = MetisAuthNamespaceSync(client_getter)
        self.calculations = MetisCalculationsNamespaceSync(client_getter)
        self.collections = MetisCollectionsNamespaceSync(client_getter)
        self.datasources = MetisDatasourcesNamespaceSync(client_getter)


class MetisAPI(MetisBase):
    """Metis API synchronous client"""

    def __init__(
        self, base_url: StrOrURL, **opts: Unpack[MetisAPIKwargs]
    ):  # pylint: disable=too-many-arguments
        """
        Initialize sync Metis client.
        Sync client is a tiny wrapper over full async client.

        **Arguments**:

        `bese_url`
        URL of Metis BFF server. `str` or `yarl.URL`.

        `headers` (Optional)
        Optional dictionary of always sent headers. Used if `session` is omitted.

        `timeout` (Optional)
        Optional integer of global timeout in seconds or `aiohttp.ClientTimeout`.
        Used if `session` is omitted.

        `client_name` (Optional)
        Optional string for user agent.
        """
        self._ns_v0 = MetisV0NamespaceSync(partial(MetisAPIAsync, base_url, **opts))

    @property
    def v0(self) -> MetisV0NamespaceSync:  # pylint: disable=invalid-name
        """Property to access the v0 namespace."""
        return self._ns_v0



"Base response DTO"

import sys

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class MetisRequestIdDTO(TypedDict):
    "Response with request id dictionary"
    reqId: str



"""Calculations endpoints namespace"""
import sys
from datetime import datetime
from functools import partial
from typing import Optional

from ..dtos import MetisCalculationDTO, MetisRequestIdDTO
from ..helpers import raise_on_metis_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence


class MetisCalculationsNamespace(BaseNamespace):
    """Calculations endpoints namespace"""

    async def cancel_event(self, calc_id: int) -> MetisRequestIdDTO:
        "Cancel calculation"
        async with await self._client.request(
            method="DELETE",
            url=self._base_url / str(calc_id),
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def cancel(self, calc_id: int) -> None:
        "Cancel calculation and wait for result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.cancel_event, calc_id)
        )

    async def create_event(
        self, data_id: int, engine: str = "dummy"
    ) -> MetisRequestIdDTO:
        "Create calculation"
        async with await self._client.request(
            method="POST",
            url=self._base_url,
            json={"dataId": data_id, "engine": engine},
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def create(
        self, data_id: int, engine: str = "dummy"
    ) -> Optional[MetisCalculationDTO]:
        "Create calculation and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.create_event, data_id, engine)
        )
        if evt["type"] == "calculations":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("createdAt", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    @raise_on_metis_error
    async def get_engines(self) -> Sequence[str]:
        "Get supported calculation engines"
        async with await self._client.request(
            method="GET",
            url=self._base_url / "engines",
            auth_required=False,
        ) as resp:
            return await resp.json()

    async def list_event(self) -> MetisRequestIdDTO:
        "List all user's calculations"
        async with await self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def list(self) -> Sequence[MetisCalculationDTO]:
        "List all user's calculations and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "calculations":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover



"""Root namespace"""

from yarl import URL

from ..client import MetisClient
from .base import BaseNamespace
from .stream import MetisStreamNamespace
from .v0 import MetisV0Namespace


class MetisRootNamespace(BaseNamespace):
    """Root namespace"""

    def __init__(
        self,
        client: MetisClient,
        base_url: URL,
        root=None,
        auth_req: bool = True,
    ) -> None:
        """Initialise the namespace."""
        super().__init__(client, base_url, self, auth_req)

    def __post_init__(self) -> None:
        self.__ns_v0 = MetisV0Namespace(self._client, self._base_url / "v0", root=self)
        self.__ns_stream = MetisStreamNamespace(
            self._client, self._base_url / "stream", root=self
        )

    @property
    def v0(self) -> "MetisV0Namespace":  # pylint: disable=invalid-name
        """Property to access the v0 namespace."""
        return self.__ns_v0

    @property
    def stream(self) -> "MetisStreamNamespace":
        """Property to access the stream namespace."""
        return self.__ns_stream



"""Main Metis API async client"""

import sys
from types import TracebackType
from typing import Optional, Type

import aiohttp
from aiohttp import ClientSession, ClientTimeout, TraceConfig
from aiohttp.hdrs import USER_AGENT
from aiohttp.typedefs import LooseHeaders, StrOrURL
from yarl import URL

from .client import MetisClient
from .const import DEFAULT_USER_AGENT
from .models import BaseAuthenticator, MetisBase
from .namespaces.root import MetisRootNamespace
from .namespaces.stream import MetisStreamNamespace
from .namespaces.v0 import MetisV0Namespace

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import List
else:  # pragma: no cover
    List = list

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired, TypedDict, Unpack
else:  # pragma: no cover
    from typing import NotRequired, TypedDict, Unpack


class MetisAPIKwargs(TypedDict):
    "MetisAPI init kwargs"
    auth: BaseAuthenticator
    headers: NotRequired[LooseHeaders]
    timeout: NotRequired[int]
    client_name: NotRequired[str]
    trace_configs: NotRequired[List[TraceConfig]]


class MetisAPIAsync(MetisBase):
    """Main Metis API async client"""

    _close_session = False
    _session: aiohttp.ClientSession

    def __init__(
        self,
        base_url: StrOrURL,
        session: Optional[ClientSession] = None,
        **opts: Unpack[MetisAPIKwargs],
    ):
        """
        Initialize async Metis client.

        **Arguments**:

        `bese_url`
        URL of Metis BFF server. `str` or `yarl.URL`.

        `session` (Optional)
        `aiohttp.ClientSession` to be used by this package.
        If you do not pass one, one will be created for you.

        `headers` (Optional)
        Optional dictionary of always sent headers. Used if `session` is omitted.

        `timeout` (Optional)
        Optional integer of global timeout in seconds or `aiohttp.ClientTimeout`.
        Used if `session` is omitted.

        `client_name` (Optional)
        Optional string for user agent.
        Used if `session` is omitted.
        """
        headers = opts.get("headers")
        if session is None:
            timeout = opts.get("timeout")
            session = aiohttp.ClientSession(
                timeout=timeout
                if isinstance(timeout, ClientTimeout)
                else ClientTimeout(total=timeout),
                headers=headers,
                trace_configs=opts.get("trace_configs"),
            )
            session.headers[USER_AGENT] = opts.get("client_name", DEFAULT_USER_AGENT)
            self._close_session = True
        self._session = session

        base_url = URL(base_url)
        if not base_url.is_absolute():
            raise TypeError("Base URL should be absolute")
        client = MetisClient(session, base_url, opts["auth"])
        self._ns_root = MetisRootNamespace(client, base_url)

    @property
    def v0(self) -> MetisV0Namespace:  # pylint: disable=invalid-name
        """Property to access the v0 namespace."""
        return self._ns_root.v0

    @property
    def stream(self) -> MetisStreamNamespace:
        """Property to access the stream namespace."""
        return self._ns_root.stream

    async def __aenter__(self) -> "MetisAPIAsync":
        """Async enter."""
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def close(self) -> None:
        "Close stream and http session"
        self.stream.close()
        if self._session and self._close_session:
            await self._session.close()



"""Data sources DTOs"""

import sys

from .base import MetisTimestampsDTO
from .collection import MetisCollectionDTO

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


class MetisDataSourceDTO(MetisTimestampsDTO):
    "Data source DTO"

    id: int
    userId: int
    userFirstName: str
    userLastName: str
    userEmail: str
    name: str
    content: str
    type: int
    collections: Sequence[MetisCollectionDTO]
    progress: int


class MetisDataSourceContentOnlyDTO(TypedDict):
    "Data source content only DTO"
    content: str



"""Base of all namespaces"""

from __future__ import annotations

from typing import TYPE_CHECKING

from yarl import URL

from ..client import MetisClient
from ..models.base import MetisBase

if TYPE_CHECKING:  # pragma: no cover
    from .root import MetisRootNamespace


class BaseNamespace(MetisBase):  # pylint: disable=too-few-public-methods
    """Used for the Metis API namespace."""

    _base_url: URL
    _client: MetisClient
    _auth_required: bool = True
    _root: MetisRootNamespace

    def __init__(
        self,
        client: MetisClient,
        base_url: URL,
        root: MetisRootNamespace,
        auth_req: bool = True,
    ) -> None:
        """Initialise the namespace."""
        self._client = client
        self._base_url = base_url
        self._auth_required = auth_req
        self._root = root
        self.__post_init__()

    def __post_init__(self) -> None:
        """Post initialisation."""



"""Collections DTO's"""
import sys
from typing import Literal, Union

from .base import MetisTimestampsDTO

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired, TypedDict
else:  # pragma: no cover
    from typing import NotRequired, TypedDict


MetisCollectionVisibility = Union[
    Literal["private"], Literal["shared"], Literal["community"]
]


class MetisCollectionTypeDTO(MetisTimestampsDTO):
    "Collection type DTO"

    id: int
    slug: str
    label: str
    flavor: str


class MetisCollectionCommonDTO(TypedDict):
    "Common fields of collection's DTOs"

    title: str
    typeId: int
    dataSources: NotRequired[Sequence[int]]
    users: NotRequired[Sequence[int]]


class MetisCollectionCreateDTO(MetisCollectionCommonDTO):
    "Collection create DTO"

    description: NotRequired[str]


class MetisCollectionDTO(MetisCollectionCommonDTO, MetisTimestampsDTO):
    "Collection DTO"

    id: int
    description: str
    visibility: MetisCollectionVisibility

    userId: int
    userFirstName: NotRequired[str]
    userLastName: NotRequired[str]
    typeSlug: NotRequired[str]
    typeLabel: NotRequired[str]
    typeFlavor: NotRequired[str]



"""User DTOs"""

import sys

from .base import MetisTimestampsDTO

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Dict
else:  # pragma: no cover
    Dict = dict

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired
else:  # pragma: no cover
    from typing import NotRequired


class MetisUserDTO(MetisTimestampsDTO):
    "User DTO"

    id: int
    firstName: NotRequired[str]
    lastName: NotRequired[str]
    email: NotRequired[str]
    emailVerified: NotRequired[bool]
    roleLabel: NotRequired[str]
    roleSlug: NotRequired[str]
    permissions: NotRequired[Dict[str, str]]
    provider: NotRequired[str]



"""Event models"""

import json
from dataclasses import dataclass
from functools import partial

from aiohttp_sse_client.client import MessageEvent

from ..dtos import (
    MetisCalculationsEventDTO,
    MetisCollectionsEventDTO,
    MetisDataSourcesEventDTO,
    MetisErrorDTO,
    MetisErrorEventDTO,
    MetisEventDTO,
    MetisPongEventDTO,
)
from ..helpers import dict_dt_from_dt_str


@dataclass(frozen=True)
class MetisMessageEvent(MessageEvent):
    "SSE Message Event"

    type: str
    message: str
    data: str
    origin: str
    last_event_id: str

    @classmethod
    def from_dto(cls, data: MessageEvent) -> "MetisMessageEvent":
        "Create model from DTO"
        return cls(
            type=data.type,
            message=data.message,
            data=data.data,
            origin=data.origin,
            last_event_id=data.last_event_id,
        )

    def to_dto(self) -> MetisEventDTO:
        "Create DTO from model"
        try:
            if self.type is None and self.message == "" and self.data == "pong":
                return MetisPongEventDTO(type="pong", data=None)
            for dto in (
                partial(MetisErrorEventDTO, type="errors"),
                partial(MetisDataSourcesEventDTO, type="datasources"),
                partial(MetisCalculationsEventDTO, type="calculations"),
                partial(MetisCollectionsEventDTO, type="collections"),
            ):
                if dto.keywords["type"] in [self.type, self.message]:
                    data = json.loads(self.data)
                    # convert date strings to datetime
                    if isinstance(data, dict) and isinstance(data.get("data"), list):
                        data["data"] = list(map(dict_dt_from_dt_str, data["data"]))
                    if isinstance(data, dict) and isinstance(data.get("types"), list):
                        data["types"] = list(map(dict_dt_from_dt_str, data["types"]))

                    return dto(data=data)
        except json.JSONDecodeError as err:
            return MetisErrorEventDTO(
                type="errors",
                data={"reqId": "", "data": [MetisErrorDTO(status=400, error=str(err))]},
            )
        message = f"Unknown event type: {str(self)}"
        return MetisErrorEventDTO(
            type="errors",
            data={"reqId": "", "data": [MetisErrorDTO(status=400, error=message)]},
        )



"""Custom exceptions for metis_client."""


from typing import Optional


class MetisException(BaseException):
    """
    This is raised when unknown exceptions occour.
    And it's used as a base for all other exceptions
    so if you want to catch all Metis related errors
    you should catch this base exception.
    """


class MetisConnectionException(MetisException):
    """This is raised when there is a connection issue with Metis."""


class MetisError(MetisException):
    """Base of all errors based on results"""

    status: int = 0
    message: Optional[str]

    def __init__(
        self, *args, status: Optional[int], message: Optional[str] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if status:
            self.status = status
        if message:
            self.message = message


class MetisNotFoundException(MetisError):
    """This is raised when the requested resource is not found."""


class MetisPayloadException(MetisError):
    """This is raised when the payload is invalid."""


class MetisPermissionException(MetisError):
    """This is raised when the user has no permission to do the requested resource."""


class MetisAuthenticationException(MetisError):
    """This is raised when we recieve an authentication issue."""


class MetisQuotaException(MetisError):
    """This is raised when quota excided."""



"""Base of DTOs"""

import sys
from datetime import datetime

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import NotRequired, TypedDict
else:  # pragma: no cover
    from typing import NotRequired, TypedDict


class MetisTimestampsDTO(TypedDict):
    "Response with timestamps"
    createdAt: NotRequired[datetime]
    updatedAt: NotRequired[datetime]



"Stream subscription"
import sys
from asyncio import CancelledError, Queue, QueueFull
from contextlib import asynccontextmanager
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

from ..dtos import MetisEventDTO, MetisRequestIdDTO
from ..helpers import raise_on_metis_error_in_event
from .base import MetisBase

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Awaitable, Callable
else:  # pragma: no cover
    from collections.abc import Awaitable, Callable

if TYPE_CHECKING:  # pragma: no cover
    from .hub import MetisHub

SubscribeCallable = Callable[[], "MetisSubscription"]
RequestIdCallable = Callable[[], Awaitable[MetisRequestIdDTO]]


@raise_on_metis_error_in_event
async def act_and_get_result_from_stream(
    sub_func: SubscribeCallable, func: RequestIdCallable
) -> MetisEventDTO:
    "Do a request and get response from stream"
    async with sub_func() as sub:
        resp = await func()
        async for msg in sub:
            data = msg.get("data")
            if isinstance(data, dict) and data.get("reqId") == resp.get("reqId"):
                return msg
    raise CancelledError  # pragma: no cover


class MetisSubscription(MetisBase):
    "Message subscription"
    hub: "MetisHub"
    queue: "Queue[MetisEventDTO]"

    def __init__(
        self,
        hub: "MetisHub",
        predicate: Optional[Callable[[MetisEventDTO], bool]] = None,
        queue_size: int = 0,
    ) -> None:
        self.hub = hub
        self.queue = Queue(maxsize=queue_size)
        self._predicate = predicate

    def put_nowait(self, message: MetisEventDTO) -> None:
        "Put message to query without wait"
        try:
            if self._predicate is None or self._predicate(message):
                self.queue.put_nowait(message)
        except QueueFull:
            self.logger.warning("Subscription's queue is full")

    async def close(self) -> None:
        "Close subscription and join message queue"
        self.hub.unsubscribe(self)
        await self.queue.join()

    def __len__(self) -> int:
        return self.queue.qsize()

    async def __aenter__(self) -> "MetisSubscription":
        self.hub.subscribe(self)
        await self.hub.wait_connected()
        return self.__aiter__()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.hub.unsubscribe(self)

    def __aiter__(self) -> "MetisSubscription":
        return self

    @asynccontextmanager
    async def _cm(self):
        try:
            yield await self.queue.get()
        finally:
            self.queue.task_done()

    async def __anext__(self) -> MetisEventDTO:
        async with self._cm() as msg:
            return msg



"""Event DTOs"""
import sys
from typing import Literal, Union

from .calculation import MetisCalculationDTO
from .collection import MetisCollectionDTO, MetisCollectionTypeDTO
from .datasource import MetisDataSourceDTO
from .error import MetisErrorDTO

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict

MetisEventType = Literal["calculations", "collections", "datasources", "errors", "pong"]


class MetisErrorEventDataDTO(TypedDict):
    "Errors event data DTO"
    reqId: str
    data: Sequence[MetisErrorDTO]


class MetisErrorEventDTO(TypedDict):
    "Errors event DTO"
    type: Literal["errors"]
    data: MetisErrorEventDataDTO


class MetisPongEventDTO(TypedDict):
    "Pong event DTO"
    type: Literal["pong"]
    data: None


class MetisDataSourcesEventDataDTO(TypedDict):
    "Data sources event data DTO"
    reqId: str
    data: Sequence[MetisDataSourceDTO]
    total: int
    types: Sequence[MetisCollectionTypeDTO]


class MetisDataSourcesEventDTO(TypedDict):
    "Data sources event DTO"
    type: Literal["datasources"]
    data: MetisDataSourcesEventDataDTO


class MetisCalculationsEventDataDTO(TypedDict):
    "Calculations event data DTO"
    reqId: str
    data: Sequence[MetisCalculationDTO]
    total: int
    types: Sequence[MetisCollectionTypeDTO]


class MetisCalculationsEventDTO(TypedDict):
    "Calculations event DTO"
    type: Literal["calculations"]
    data: MetisCalculationsEventDataDTO


class MetisCollectionsEventDataDTO(TypedDict):
    "Collections event data DTO"
    reqId: str
    data: Sequence[MetisCollectionDTO]
    total: int
    types: Sequence[MetisCollectionTypeDTO]


class MetisCollectionsEventDTO(TypedDict):
    "Collections event DTO"
    type: Literal["collections"]
    data: MetisCollectionsEventDataDTO


MetisEventDTO = Union[
    MetisCalculationsEventDTO,
    MetisCollectionsEventDTO,
    MetisDataSourcesEventDTO,
    MetisErrorEventDTO,
    MetisPongEventDTO,
]



"""Constants for metis_client."""

from __future__ import annotations

from logging import Logger, getLogger
from typing import Literal, Union

PROJECT_VERSION = "0.0.3"
PROJECT_NAME = "metis_client"

# This is the default user agent,
# but it is adviced to use your own when building out your application
DEFAULT_USER_AGENT = f"metis_client/{PROJECT_VERSION}"


LOGGER: Logger = getLogger(PROJECT_NAME)

HttpMethods = Union[
    Literal["CONNECT"],
    Literal["HEAD"],
    Literal["GET"],
    Literal["DELETE"],
    Literal["OPTIONS"],
    Literal["PATCH"],
    Literal["POST"],
    Literal["PUT"],
    Literal["TRACE"],
]
